#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
"""
String Methods
"""


def binary_string(number):
    """Number to binary string

    :param number: some number (an integer) to turn into a binary string
    :return: Some string which is the binary string
    """
    return bin(number)[2:]


def string_score(strang):
    """Sum of letter values where a==1 and z == 26

    :param strang: string to be scored
    :type strang: str
    :returns: -> score of the string
    :rtype: int

    .. doctest::python

        >>> string_score('me')
        18
        >>> string_score('poooood')
        95
        >>> string_score('gregory')
        95

    """
    return sum((ord(character) - 96 for character in strang.lower()))


def is_palindrome(string):
    """True a string is a palindrome; False if string is not a palindrome.

    :param string: 

    .. doctest::python

        >>> is_palindrome("racecar")
        True
        >>> is_palindrome("greg")
        False

    """
    return all(
        character == string[-index - 1] for index, character in enumerate(string)
    )




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from typing import Any
from typing import Generator
from typing import Iterable
from typing import List
from typing import Sequence
from typing import Union

Flint = Union[int, float]

def partitions_gen(numero: int, min_p: int = ..., max_p: int = ...) -> Generator: ...
def rfactorial(n: int): ...
def radians_2_degrees(rads: Flint): ...
def degrees_2_radians(degs: Flint): ...
def power_mod(number: Flint, exponent: Flint, mod: Flint): ...
def divisors_gen(n: int) -> Generator: ...
def gcd_it(a: int, b: int): ...
def gcd_r(a: int, b: int): ...
def reverse(n: int): ...
def fib_r(n: int): ...
def expo(d: Any, n: Any): ...
def pytriple_gen(max_c: int) -> Generator: ...
def repermutations(toop: tuple) -> int: ...
def disjoint(a: Iterable, b: Iterable) -> bool: ...
def n_choose_r(n: int, r: int) -> int: ...
def pytriple_gen_2() -> None: ...
def get_pythag_triple(real: Any, imag: Any): ...

class Trigon:
    pt1: Any = ...
    pt2: Any = ...
    pt3: Any = ...
    def __init__(self, pt1: Any, pt2: Any, pt3: Any) -> None: ...
    @classmethod
    def from_points(cls, pts: List[Any]): ...
    def __str__(self): ...
    def __contains__(self, point: Any): ...
    def inner_triangles(self, point: Any): ...
    def is_perimeter_point(self, point: Any): ...
    def points(self): ...
    def contains_origin(self): ...
    def area(self): ...
    @staticmethod
    def area_from_points(pt1: Any, pt2: Any, pt3: Any): ...

class Vuple(tuple):
    Voops = Union[tuple, int, float, Sequence]
    def __new__(cls, *args: Any): ...
    def __gt__(self, other: Any): ...
    def __eq__(self, other: Any): ...
    def __add__(self, k: Voops): ...
    def __sub__(self, k: Voops): ...
    def __mul__(self, k: Union[tuple, int, float]): ...
    def __imul__(self, k: Union[tuple, int, float]): ...
    def _mul_scalar(self, k: Flint): ...
    def __truediv__(self, k: Voops): ...
    def _truediv_scalar(self, k: Flint): ...
    def __itruediv__(self, k: Voops): ...
    def __floordiv__(self, k: Voops): ...
    def __ifloordiv__(self, k: Voops): ...
    def _floordiv_scalar_int(self, k: Flint): ...
    def normalize(self): ...
    def get_mag_sqrd(self): ...
    def get_mag(self): ...
    def is_disjoint(self, them: Any): ...
    def product(self): ...
    @staticmethod
    def unit_vuple(voop: Union[Vuple, tuple]): ...
    @staticmethod
    def mag_sqrd(voop: Union[Vuple, tuple, List[Union[int, float]]]): ...
    @staticmethod
    def mag(voop: Any): ...
    @staticmethod
    def dot(a: Any, b: Any): ...
    @staticmethod
    def cross(v1: Any, v2: Any): ...
    @staticmethod
    def angle(
        v1: Union[Vuple, tuple], v2: Union[Vuple, tuple], radians: bool = ...
    ): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from __future__ import division
from __future__ import print_function

import collections
from bisect import bisect
from bisect import bisect_right
from itertools import count
from math import sqrt
from sys import version_info

from pupy.decorations import cash_it
from pupy.maths import divisors_gen

try:
    range
except NameError:
    range = xrange


def prime_gen(plim=0, kprimes=None):
    """Infinite (within reason) prime number generator
    
    My big modification is the pdiv_dictionary() function that recreats the
    dictionary of divisors so that you can continue to generate prime numbers
    from a (sorted) list of prime numbers.
    
    Based on:
        eratosthenes by David Eppstein, UC Irvine, 28 Feb 2002
        http://code.activestate.com/recipes/117119/ and the thread at the url

    :param plim: prime_limit; default=0 makes for an infinite generator
    :type plim: int
    :param kprimes: known_primes as an iterable (Default value = None)
    :type kprimes: iter

    .. doctest::

        >>> list(prime_gen(50))
        [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47]
        >>> list(prime_gen(10))
        [2, 3, 5, 7]

    """

    if kprimes is None:
        kprimes = [2, 3, 5, 7, 11]

    def _dictionary():
        """Recreates the prime divisors dictionary used by the generator"""
        div_dict = {}
        for pdiv in kprimes:
            multiple = kprimes[-1] // pdiv * pdiv
            if multiple % 2 == 0:
                multiple += pdiv
            else:
                multiple += 2 * pdiv
            while multiple in div_dict:
                multiple += pdiv * 2
            div_dict[multiple] = pdiv
        return div_dict

    # [1]
    # See if the upper bound is greater than the known primes
    if 0 < plim <= kprimes[-1]:
        for p in kprimes:
            if p <= plim:
                yield p
        return  # return bc we are done

    # [2]
    # Recreate the prime divisibility dictionary using kprimes;
    # Set start and yield first 4 primes
    divz = _dictionary()
    start = kprimes[-1] + 2  # max prime + 2 (make sure it is odd)
    if start == 13:
        yield 2
        yield 3
        yield 5
        yield 7
        yield 11
    # use count or range depending on if generator is infinite
    it = count(start, 2) if plim == 0 else range(start, plim, 2)
    for num in it:
        prime_div = divz.pop(num, None)
        if prime_div:
            multiple = (2 * prime_div) + num
            while multiple in divz:
                multiple += 2 * prime_div
            divz[multiple] = prime_div
        else:
            divz[num * num] = num
            yield num


def prime_factorization_gen(n):
    """generates all numbers in the prime factorization of n

    :param n: number to be factored
    :type n: int

    .. doctest::

        >>> list(prime_factorization_gen(12))
        [2, 2, 3]
        >>> list(prime_factorization_gen(16))
        [2, 2, 2, 2]

    """
    for factor in prime_factors_gen(n):
        if n <= 1:
            break
        while n % factor == 0:
            n //= factor
            yield factor


def prime_factors_gen(n):
    """prime factors generator

    :param n: number to be factorized
    :type n: int

    .. doctest:: python

        >>> list(prime_factors_gen(12))
        [2, 3]
        >>> list(prime_factors_gen(16))
        [2]

    """
    return (p for p in divisors_gen(n) if is_prime(p))


@cash_it
def is_prime(number):
    """Checks if a number is prime

    :param number: number to check if is prime
    :type number: int
    :returns: -> True if number is prime
    :rtype: bool

    .. doctest:: python

        >>> is_prime(37)
        True
        >>> is_prime(100)
        False
        >>> is_prime(89)
        True

    """
    if number == 2 or number == 3:
        return True
    if number < 2 or number % 2 == 0:
        return False
    if number < 9:
        return True
    if number % 3 == 0:
        return False
    for step in range(5, int(sqrt(number)) + 1, 6):
        if step >= number:
            break
        if number % step == 0:
            return False
        if number % (step + 2) == 0:
            return False
    return True


# class OctopusPrime(list):
class OctopusPrime(collections.MutableSequence):
    """OctopusPrime, the 8-leg autobot, here to help you find PRIMES"""

    def __init__(self, plim=100):
        # list.__init__(self, list(prime_gen(plim=plim)))
        # super(OctopusPrime, self).__init__()
        p = [
            2,
            3,
            5,
            7,
            11,
            13,
            17,
            19,
            23,
            29,
            31,
            37,
            41,
            43,
            47,
            53,
            59,
            61,
            67,
            71,
            73,
            79,
            83,
            89,
            97,
        ]
        if plim == 100:
            self._list = p[:]
        elif plim < 100:
            self._list = list(filter(lambda n: n <= plim, p))
        else:
            self._list = list(prime_gen(plim, p))
        self.max_loaded = self._list[-1]

    def _transform(self, n=None):
        """TRANSFORM / grow the list

        :param n:  (Default value = None)

        """
        n = n if n is not None else self._list[-1] * 10
        self._list.extend(list(prime_gen(plim=n, kprimes=self._list)))

    def primes_below(self, upper_bound):
        """Lists primes, p, such that  p < upper_bound

        :param upper_bound: exclusive upper bound
        :type upper_bound: int
        :returns: -> primes less than upper_bound
        :rtype: list

        """
        return self.primes_between(1, upper_bound)

    def primes_between(self, lower_bound, upper_bound):
        """Lists primes, p, such that, lower_bound < p < upper_bound

        :param lower_bound: exclusive lower bound
        :type lower_bound: int
        :param upper_bound: exclusive upper bound
        :type upper_bound: int
        :returns: -> primes between lower_bound and upper_bound
        :rtype: list

        """
        if upper_bound > self[-1]:
            self._transform(upper_bound)
        return self[bisect_right(self, lower_bound) : bisect(self, upper_bound)]

    def __len__(self):
        return len(self._list)

    def __getitem__(self, i):
        return self._list[i]

    def __delitem__(self, i):
        del self._list[i]

    def __setitem__(self, key, value):
        self._list[key] = value

    def insert(self, index, object):
        """

        :param index: 
        :param object: 

        """
        self._list.insert(index, object)

    def __str__(self):
        return str(self._list)

    def __repr__(self):
        return str(self._list)


if __name__ == "__main__":
    from doctest import testmod

    testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from __future__ import print_function
from __future__ import with_statement

from codecs import getwriter
from datetime import datetime
from io import open
from itertools import count
from os import path
from os import remove
from os import utime
from time import sleep

from pupy.decorations import mkdirs

try:
    from ujson import dump
    from ujson import load
except ModuleNotFoundError as e:
    from json import dump
    from json import load

def timestamp():
    """Time stamp string w/ format yyyy-mm-ddTHH-MM-SS

    :return: timestamp string
    """
    """Time stamp string w/ format yyyy-mm-ddTHH-MM-SS"""
    return datetime.now().strftime("%Y-%m-%dT%H-%M-%S")

def safe_path(filepath):
    """

    :param filepath: 

    """
    if path.exists(filepath):
        f_bn, f_ext = path.splitext(filepath)
        for n in count(1):
            safe_save_path = f_bn + "_({}).".format(str(n)) + f_ext
            if not path.exists(safe_save_path):
                return safe_save_path
    return filepath

def ensure_save(filepath, n=0):
    try:
        assert path.exists(filepath)
        return True
    except AssertionError:
        sleep(1)
        if n > 5:
            return False
        return ensure_save(filepath, n + 1)

@mkdirs
def savings(filepath, string, clobber=True):
    """Save s(tring) to filepath as txt file

    :param filepath: Filepath save location
    :param string: File as a string to be saved
    :param clobber: Save over a file if the filepath exists
    :return: None? what do you want? confirmation?

    """
    if not clobber and path.exists(filepath):
        filepath = safe_path(filepath)
    elif path.exists(filepath):
        remove(filepath)
    try:
        with open(safe_path(filepath) if clobber else filepath, "wb") as file:
            file.write(string.encode("utf-8"))
    except Exception as e:
        raise e

@mkdirs
def loads(filepath):
    """Load a (txt) file as a string

    :param filepath: return:

    """
    try:
        with open(filepath, "r", encoding="utf-8") as f:
            return f.read()
    except UnicodeDecodeError as e:
        with open(filepath, "r", encoding="latin2") as f:
            return f.read()

@mkdirs
def sjson(filepath, data, min=False):
    """Save json-serial-ize-able data to a specific filepath.

    :param filepath: save filepath
    :param data: json ready data
    :param min: minified format flag
    :return: None
    """
    if type(data) == dict and any(type(val) == bytes for val in data.values()):
        data = {k: str(v, encoding="utf-8") for k, v in data.items()}
    with open(filepath, "wb") as jsonfile:
        if min:
            dump(data, getwriter("utf-8")(jsonfile), ensure_ascii=False)
        else:
            dump(
                data,
                getwriter("utf-8")(jsonfile),
                indent=4,
                sort_keys=True,
                ensure_ascii=False,
                )

def save_jasm(filepath, data, min=False):
    """Alias for sjson (which stands for 'save-json')"""
    return sjson(filepath, data, min)

def sjasm(filepath, data, min=False):
    """Alias for sjson (which stands for 'save-json')"""
    return sjson(filepath, data, min)

def spak(filepath, data, min=False):
    """Alias for sjson (which stands for 'save-json')"""
    return sjson(filepath, data, min)

@mkdirs
def ljson(filepath):
    """Load a json file given a filepath and return the file-data

    :param filepath: path to the jasm file you want to load
    :return: Loaded file contents
    """

    with open(filepath) as infile:
        return load(infile)

def load_jasm(filepath):
    """Alias for ljson (which stands for 'load-json')"""
    return ljson(filepath)

def ljasm(filepath):
    """Alias for ljson (which stands for 'load-json')"""
    return ljson(filepath)

def lpak(filepath):
    """Alias for ljson (which stands for 'load-json')"""
    return ljson(filepath)

@mkdirs
def touch(filepath):
    """Touches a file just like touch on the command line"""
    with open(filepath, 'a'):
        utime(filepath, None)

class Jasm(object):
    """Jasm the Grundle Bug"""

    def __init__(self, path):
        self.file_path = path

    def __call__(self, funk):
        """Json saving and loading"""
        fp = self.file_path

        def savings_n_loads(*args, **kwargs):
            """Jasm funk (w)rapper

            """
            if len(args) == 0:
                save_key = "None"
            else:
                save_key = str((args, kwargs.items()))
            try:
                dat_data = ljson(fp)
            except IOError:
                dat_data = {}
            except ValueError:
                dat_data = {}

            if save_key not in dat_data:
                ret_val = funk(*args, **kwargs)
                dat_data[save_key] = ret_val
                sjson(fp, dat_data)
            return dat_data[save_key]

        return savings_n_loads



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

def binary_string(number: int) -> str: ...
def string_score(string: str) -> int: ...
def is_palindrome(string: str) -> bool: ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any
from typing import Generator
from typing import Iterable
from typing import List
from typing import Union

def chunks(list: List[Any], chunk_size: int) -> Generator: ...
def is_permutation(a: Union[Iterable[Any], int], b: Union[Iterable[Any], int]): ...
def rotate(rlist: Any, rn: int = ..., left_rotate: bool = ...): ...
def rotations_gen(rlist: Union[List[Any], tuple]) -> None: ...
def digits_list(number: int): ...
def int_from_digits(dlist: Union[List[int], tuple, Iterable[int]]): ...
def iter_product(l: Union[List[int], tuple, Iterable[int]]): ...
def files_gen(dirpath:str, abs:bool): ...
def dirs_gen(dirpath:str, abs:bool): ...




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from typing import Any
from typing import Optional

class SodokuError(ValueError):
    message: str = ...
    def __init__(
        self, message: str, row: Optional[int] = ..., col: Optional[int] = ...
    ) -> None: ...

class Sodoku:
    is_solved: bool = ...
    board: Any = ...
    def __init__(self, board: Any) -> None: ...
    def solve(self) -> None: ...
    def euler_096_three_digit_number(self): ...
    @staticmethod
    def first_unknown(d: Any): ...
    @staticmethod
    def unsolvable(rcbd: Any): ...
    @staticmethod
    def check_unsolvable(d: Any): ...
    @staticmethod
    def update_dictionary(d: Any): ...
    @staticmethod
    def reduce_dictionary(d: Any): ...
    @staticmethod
    def hasdup(d: Any): ...
    def get_oneline_str(self): ...
    @staticmethod
    def neighbors(index: Any, size: int = ...): ...
    @staticmethod
    def irow(n: Any, bsize: int = ...): ...
    @staticmethod
    def icolumn(n: Any, bsize: int = ...): ...
    @staticmethod
    def ibox(box_r: Any, box_c: Any, bsize: int = ...): ...
    @staticmethod
    def box_box(index: Any, bsize: int = ...): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from collections import MutableSequence
from typing import Any
from typing import List

def prime_gen(plim: int = ..., kprimes: List[int] = ...): ...
def prime_factorization_gen(n: int) -> None: ...
def prime_factors_gen(n: int): ...
def is_prime(number: int): ...

class OctopusPrime(MutableSequence):
    max_loaded: int = ...
    def __init__(self, plim: int = ...) -> None:
        self._list = None
        ...
    def __len__(self) -> int: ...
    def __delitem__(self, key) -> None: ...
    def __getitem__(self, key) -> Any: ...
    def __setitem__(self, key): ...
    def _transform(self, n: int = ...) -> None: ...
    def primes_below(self, upper_bound: int): ...
    def primes_between(self, lower_bound: int, upper_bound: int): ...
    def insert(self, key): ...
